#include <BluetoothSerial.h>
#include "ELMduino.h"
#include <SSD1283A.h>
#include "dsdigital.c"                // custom made DS-digital font
#include <U8g2_for_Adafruit_GFX.h>
#include "jaguar.h"                   //defines the jaguar logo file
#include "ezButton.h"

BluetoothSerial SerialBT;
#define ELM_PORT SerialBT
//uint8_t address[6]  = {0x86, 0xDC, 0x49, 0xAB, 0xF7, 0xF1};        // old MAC
uint8_t address[6]  = {0xDC, 0x0D, 0x30, 0x48, 0x8D, 0x60};        // define the ELM327 MAC address here
// mine: (DC:0D:30:48:8D:60) yours could be different
ELM327 myELM327;

SSD1283A screen(/*CS=5*/ SS, /*DC=*/ 17, /*RST=*/ 16, /*LED=*/ 4); // pins to ESP32: hardware spi,cs,cd,reset, led
U8G2_FOR_ADAFRUIT_GFX gfx;

GFXcanvas16T<130, 130> canvas;

uint32_t speedo = 0;
uint32_t Coolant = 0;                 // added to ELMduino.cpp and ELMDuino.h [float Coolant();]
float    volt = 0.0;                  // added to ELMduino.cpp and ELMDuino.h [float volt();]
uint32_t driving = 0;                 // added to ELMduino.cpp and ELMDuino.h [float driving();]
uint32_t Intake = 0;                  // added to ELMduino.cpp and ELMDuino.h [float Intake();]
uint32_t Ambient = 0;                 // added to ELMduino.cpp and ELMDuino.h [float Ambient();]
uint32_t Rpm = 0;

const byte pResistor = 2;             // Photo resistor GL5528 connected to ESP32 analog pin GPIO 2, divided with resistor 8.2KΩ
int bright = 0;                       // Store brightness value from photo resistor (0-4095)
int lux = 0;                          // recalculated brightness value (0-255)
const int MIN_LIGHT = 0;              // Minimum expected brightness
const int MAX_LIGHT = 4095;           // Maximum expected brightness
int pwm_pin = 4;                      // pin from LED (display) to ESP32 GPIO 4
const int freq = 5000;                // PWM frequency
const int ledChannel = 0;
const int resolution = 8;

int passFlag1 = 0;                    // refresh static data only once
int passFlag2 = 0;
int passFlag3 = 0;
int passFlag4 = 0;
int passFlag5 = 0;
int passFlag6 = 0;
int passFlag7 = 0;

float lastCoolant = 1000;
float lastSpeedo = 1000;
float lastVolt = 0.0;
float lastDriving = 10000;
float lastAmbient = 1000;
float lastIntake = 1000;
float lastRpm = 10000;

#define BLACK   0x0000                // define custom color <RGB565>, visit: http://www.barth-dev.de/online/rgb565-color-picker/
#define WHITE   0xFFFF
#define ORANGE  0xFA60
#define LIME    0xAFEB

uint8_t rotation = 2;                 // rotate the display (1 to 4)

const int SHORT_PRESS_TIME = 500; // 500 milliseconds
const int LONG_PRESS_TIME  = 500; // 500 milliseconds

ezButton button(34);                  // button pin to ESP32 pin GPiO 34 (D34)
unsigned long pressedTime  = 0;
unsigned long releasedTime = 0;
bool isPressing = false;
bool isLongDetected = false;
int buttonPushCounter = 0;            // counter for the number of button presses
int buttonState = 0;                  // current state of the button
int lastButtonState = 0;              // previous state of the button
int delayMe = 300;                    // refresh the screen in milliseconds
unsigned long time_now = 0;

//_____________________________________________________________

void deBounce() {           //debounce the button
  button.loop(); // MUST call the loop() function first
  if (button.isPressed()) {
    pressedTime = millis();
    isPressing = true;
    isLongDetected = false;
  }
  if (button.isReleased()) {
    isPressing = false;
    releasedTime = millis();

    long pressDuration = releasedTime - pressedTime;

    if ( pressDuration < SHORT_PRESS_TIME ) {
      buttonState = 1;
      if (buttonState != lastButtonState) {
        buttonPushCounter++;
      }
    }
  }

  if (isPressing == true && isLongDetected == false) {
    long pressDuration = millis() - pressedTime;

    if ( pressDuration > LONG_PRESS_TIME ) {
      buttonPushCounter = 1;                                  // change to your long button press preference (0 to 6)
      isLongDetected = true;
    }
  }
  if (buttonPushCounter == 7) {
    buttonPushCounter = 0;
  }
}

//_____________________________________________________________

void loop() {

  deBounce();
  if (millis() >= time_now + delayMe) {               //delayMe = 300 milliseconds
    time_now += delayMe;
    brightness();

    if (buttonPushCounter == 0) {
      showCoolant();
    }
    else if (buttonPushCounter == 1) {
      showSpeedo();
    }
    else if (buttonPushCounter == 2) {
      showVolt();
    }
    else if (buttonPushCounter == 3) {
      showDriving();
    }
    else if (buttonPushCounter == 4) {
      showAmbient();
    }
    else if (buttonPushCounter == 5) {
      showIntake();
    }
    else if (buttonPushCounter == 6) {
      showRpm();
    }
  }
}

//_____________________________________________________________

void show_canvas_on_screen()
{
  screen.drawRGBBitmap(0, 0, canvas.getBuffer(), canvas.width(), canvas.height());
}

//_____________________________________________________________

void setup()
{
  delay(200);
  ledcSetup(ledChannel, freq, resolution);
  pinMode(pwm_pin, OUTPUT);
  ledcAttachPin(pwm_pin, ledChannel);
  button.setDebounceTime(20); // set debounce time to 50 milliseconds
  pinMode(pResistor, INPUT);
  brightness();
  screen.init();
  gfx.begin(canvas);
  screen.setRotation(rotation);
  delay(200);
  canvas.fillScreen(BLACK);
  show_canvas_on_screen();

  canvas.fillScreen(BLACK);
  //canvas.drawBitmap(2, 35, jaguar1, 126, 60, LIME);      // uncomment if you want this logo
  //canvas.drawBitmap(5, 5, jaguar2, 120, 120, WHITE);     // uncomment if you want this logo
  canvas.drawBitmap(5, 31, jaguar3, 120, 68, LIME);        // comment if you want the other logo
  show_canvas_on_screen();
  delay(3000);
  canvas.fillScreen(BLACK);
  //canvas.drawBitmap(0, 0, X150, 130, 130, WHITE);        // uncomment if you want this logo
  canvas.drawBitmap(0, 23, X150_2, 130, 85, WHITE);         // comment if you want the other logo
  show_canvas_on_screen();
  delay(3000);

  ELM_PORT.begin("OBDII", true);
  delay(500);

  canvas.fillScreen(BLACK);
  canvas.setTextColor(WHITE);
  canvas.setTextSize(2);
  canvas.setCursor(3, 3);
  canvas.println(F("Attempting"));
  canvas.println(F("to connect"));
  canvas.println(F("to OBD2..."));
  show_canvas_on_screen();

  if (!ELM_PORT.connect(address))
  {
    canvas.fillScreen(BLACK);
    canvas.setTextColor(ORANGE);
    canvas.setTextSize(2);
    canvas.setCursor(3, 3);
    canvas.println(F("Couldn't"));
    canvas.println(F("connect to"));
    canvas.println(F("OBD2!"));
    show_canvas_on_screen();
  }

  if (!myELM327.begin(ELM_PORT, '6'))                  // '6' refers to OBD protocol ISO_15765_11_BIT_500_KBAUD in ELMDuino.h
  {
    canvas.fillScreen(BLACK);
    canvas.setTextColor(ORANGE);
    canvas.setCursor(1, 3);
    canvas.println(F("Couldn't"));
    canvas.println(F("talk with"));
    canvas.println(F("OBD2!"));
    show_canvas_on_screen();
    //while (1);
    delay(3000);
    canvas.fillScreen(BLACK);
    canvas.setTextColor(LIME);
    canvas.setCursor(15, 60);
    canvas.println(F("Retrying"));
    show_canvas_on_screen();
    delay(3000);
    setup();
  }
  canvas.fillScreen(BLACK);
  canvas.setTextColor(LIME);
  canvas.setCursor(3, 3);
  canvas.println(F("Connected"));
  canvas.println(F("to OBD2!"));
  show_canvas_on_screen();
  delay(3000);
  canvas.fillScreen(BLACK);
  show_canvas_on_screen();
}

//_____________________________________________________________

void showCoolant()
{ float tempCoolant = myELM327.Coolant();
  if (myELM327.status == ELM_SUCCESS)
  {
    Coolant = (uint32_t)tempCoolant;
    // Coolant = round((Coolant * 9 / 5) + 32);   //uncomment if you use Fahrenheit

  }
  else {
    elmError();
  }
  if (passFlag1 == 0) {                    //draw the static text only once
    canvas.fillScreen(BLACK);
    canvas.setTextColor(LIME);
    canvas.setTextSize(2);
    canvas.setCursor(3, 90);
    canvas.println(F("Coolant  C"));            // change C to F if you use Fahrenheit
    canvas.drawCircle(104, 92, 3, LIME);      // comment out if you use Fahrenheit
    show_canvas_on_screen();
    passFlag1++;
    passFlag7 = 0;
  }
  if (Coolant != lastCoolant) {            // prevent display refresh when the same value is read
    gfx.setFont(dsdigital_54);
    gfx.setCursor(9, 58);
    gfx.setForegroundColor(WHITE);
    canvas.fillRect(0, 12, 130, 46, BLACK);  // x start, y start, x width, y height
    gfx.println(String(Coolant));
    show_canvas_on_screen();
  }
  lastCoolant = Coolant;
  lastRpm = 10000;
}

//_____________________________________________________________

void showSpeedo()
{ float tempSPEEDO = myELM327.speedo();
  if (myELM327.status == ELM_SUCCESS)
  {
    speedo = (uint32_t)tempSPEEDO;
    // speedo = round(speedo * 0.6213711922);        //uncomment if you use MPH
  }
  else {
    elmError();
  }
  if (passFlag2 == 0) {
    canvas.fillScreen(BLACK);
    canvas.setTextColor(LIME);
    canvas.setTextSize(3);
    canvas.setCursor(25, 90);
    canvas.println(F("KM/H"));                   // change "KM/H" to "MPH" if you use miles
    show_canvas_on_screen();
    passFlag2++;
    passFlag1 = 0;
  }
  if (speedo != lastSpeedo) {
    gfx.setFont(dsdigital_54);
    gfx.setCursor(9, 58);
    gfx.setForegroundColor(WHITE);
    canvas.fillRect(13, 12, 103, 46, BLACK);
    gfx.println(String(speedo));
    show_canvas_on_screen();
  }
  lastSpeedo = speedo;
  lastCoolant = 1000;
}

//_____________________________________________________________

void showVolt()
{ float tempVOLT = myELM327.volt();
  if (myELM327.status == ELM_SUCCESS)
  {
    volt = (uint32_t)tempVOLT;
    volt = volt / 1000;
    volt = round(volt * 10);
    volt = volt / 10;
  }
  else {
    elmError();
  }
  if (passFlag3 == 0) {
    canvas.fillScreen(BLACK);
    canvas.setTextColor(LIME);
    canvas.setTextSize(3);
    canvas.setCursor(2, 90);
    canvas.println(F("Voltage"));
    show_canvas_on_screen();
    passFlag3++;
    passFlag2 = 0;
  }
  if (volt != lastVolt) {
    gfx.setFont(dsdigital_54);
    gfx.setCursor(3, 58);
    gfx.setForegroundColor(WHITE);
    canvas.fillRect(10, 12, 97, 46, BLACK);
    gfx.println(volt, 1);
    show_canvas_on_screen();
  }
  lastVolt = volt;
  lastSpeedo = 1000;
}

//_____________________________________________________________

void showDriving()
{ float tempDRIVE = myELM327.driving();
  if (myELM327.status == ELM_SUCCESS)
  {
    driving = (uint32_t)tempDRIVE;
  }
  else {
    elmError();
  }
  if (passFlag4 == 0) {
    canvas.fillScreen(BLACK);
    canvas.setTextColor(LIME);
    canvas.setTextSize(2);
    canvas.setCursor(20, 70);
    canvas.println(F("Minutes"));
    canvas.setCursor(32, 87);
    canvas.println(F("since"));
    canvas.setCursor(32, 104);
    canvas.println(F("start"));
    show_canvas_on_screen();
    passFlag4++;
    passFlag3 = 0;
  }
  if (driving != lastDriving) {
    gfx.setFont(dsdigital_54);
    gfx.setCursor(9, 58);
    gfx.setForegroundColor(WHITE);
    canvas.fillRect(13, 12, 103, 46, BLACK);
    gfx.println(String(driving));
    show_canvas_on_screen();
  }
  lastDriving = driving;
  lastVolt = 0.0;
}

//_____________________________________________________________

void showAmbient()
{ float tempAMBIENT = myELM327.Ambient();
  if (myELM327.status == ELM_SUCCESS)
  {
    Ambient = (uint32_t)tempAMBIENT;
    // Ambient = round((Ambient * 9 / 5) + 32);  //uncomment if you use Fahrenheit
  }
  else {
    elmError();
  }
  if (passFlag5 == 0) {
    canvas.fillScreen(BLACK);
    canvas.setTextColor(LIME);
    canvas.setTextSize(2);
    canvas.setCursor(3, 90);
    canvas.println(F("Outside  C"));                  // change C to F if you use Fahrenheit
    canvas.drawCircle(105, 92, 3, LIME);            // comment out if you use Fahrenheit
    show_canvas_on_screen();
    passFlag5++;
    passFlag4 = 0;
  }
  if (Ambient != lastAmbient) {
    gfx.setFont(dsdigital_54);
    gfx.setCursor(9, 58);
    gfx.setForegroundColor(WHITE);
    canvas.fillRect(13, 12, 103, 46, BLACK);
    gfx.println(String(Ambient));
    show_canvas_on_screen();
  }
  lastAmbient = Ambient;
  lastDriving = 10000;
}

//_____________________________________________________________

void showIntake()
{ float tempINTAKE = myELM327.Intake();
  if (myELM327.status == ELM_SUCCESS)
  {
    Intake = (uint32_t)tempINTAKE;
    // Intake = round((Intake * 9 / 5) + 32);    //uncomment if you use Fahrenheit
  }
  else {
    elmError();
  }
  if (passFlag6 == 0) {
    canvas.fillScreen(BLACK);
    canvas.setTextColor(LIME);
    canvas.setTextSize(2);
    canvas.setCursor(3, 90);
    canvas.println(F("Intake   C"));                  // change C to F if you use Fahrenheit
    canvas.drawCircle(105, 92, 3, LIME);            // comment out if you use Fahrenheit
    show_canvas_on_screen();
    passFlag6++;
    passFlag5 = 0;
  }
  if (Intake != lastIntake) {
    gfx.setFont(dsdigital_54);
    gfx.setCursor(9, 58);
    gfx.setForegroundColor(WHITE);
    canvas.fillRect(13, 12, 103, 46, BLACK);
    gfx.println(String(Intake));
    show_canvas_on_screen();
  }
  lastIntake = Intake;
  lastAmbient = 1000;
}

//_____________________________________________________________

void showRpm() {
  float tempRPM = myELM327.Rpm();
  if (myELM327.status == ELM_SUCCESS)
  {
    Rpm = (uint32_t)tempRPM;

  }
  else {
    elmError();
  }
  if (passFlag7 == 0) {                    //draw the static text only once
    canvas.fillScreen(BLACK);
    canvas.setTextColor(LIME);
    canvas.setTextSize(3);
    canvas.setCursor(35, 90);
    canvas.println(F("RPM"));
    show_canvas_on_screen();
    passFlag7++;
    passFlag6 = 0;
  }
  if (Rpm != lastRpm) {            // prevent display refresh when the same value is read
    gfx.setFont(dsdigital_47);
    gfx.setCursor(3, 58);
    gfx.setForegroundColor(WHITE);
    canvas.fillRect(0, 12, 130, 46, BLACK);  // x start, y start, x width, y height
    gfx.println(String(Rpm));
    show_canvas_on_screen();
  }
  lastRpm = Rpm;
  lastIntake = 1000;
}

//_____________________________________________________________

void elmError() {

  canvas.fillScreen(BLACK);
  canvas.setTextColor(LIME);
  canvas.setTextSize(2);
  canvas.setCursor(0, 20);
  canvas.println(F(" If your"));
  canvas.println(F(" engine is"));
  canvas.println(F(" running:"));
  show_canvas_on_screen();
  delay(5000);
  canvas.fillScreen(BLACK);
  canvas.setTextColor(ORANGE);
  canvas.setTextSize(3);
  canvas.setCursor(3, 3);
  canvas.println(F("Error:"));
  canvas.setCursor(40, 50);
  canvas.setTextSize(5);
  canvas.println(myELM327.status);
  show_canvas_on_screen();
  delay(4000);
  canvas.fillScreen(BLACK);
  canvas.setTextColor(LIME);
  canvas.setCursor(0, 20);
  canvas.setTextSize(2);
  canvas.println(F(" Or Start"));
  canvas.println(F(" Your Car"));
  show_canvas_on_screen();
  delay(4000);
  canvas.fillScreen(BLACK);
  passFlag1 = 0;
  passFlag2 = 0;
  passFlag3 = 0;
  passFlag4 = 0;
  passFlag5 = 0;
  passFlag6 = 0;
  passFlag7 = 0;
  lastCoolant = 1000;
  lastSpeedo = 1000;
  lastVolt = 0.0;
  lastDriving = 10000;
  lastAmbient = 1000;
  lastIntake = 1000;
  lastRpm = 10000;
  retry();
}

//_____________________________________________________________

void retry() {
  ELM_PORT.begin("OBDII", true);
  delay(200);
  brightness();

  canvas.fillScreen(BLACK);
  canvas.setTextColor(WHITE);
  canvas.setTextSize(2);
  canvas.setCursor(3, 3);
  canvas.println(F("Attempting"));
  canvas.println(F("to connect"));
  canvas.println(F("to OBD2..."));
  show_canvas_on_screen();

  if (!ELM_PORT.connect(address))
  {
    canvas.fillScreen(BLACK);
    canvas.setTextColor(ORANGE);
    canvas.setTextSize(2);
    canvas.setCursor(3, 3);
    canvas.println(F("Couldn't"));
    canvas.println(F("connect to"));
    canvas.println(F("OBD2!"));
    show_canvas_on_screen();
  }

  if (!myELM327.begin(ELM_PORT, '6'))                  // '6' refers to OBD protocol ISO_15765_11_BIT_500_KBAUD in ELMDuino.h
  {
    canvas.fillScreen(BLACK);
    canvas.setTextColor(ORANGE);
    canvas.setCursor(1, 3);
    canvas.println(F("Couldn't"));
    canvas.println(F("talk with"));
    canvas.println(F("OBD2!"));
    show_canvas_on_screen();
    //while (1);
    delay(1000);
    canvas.fillScreen(BLACK);
    canvas.setTextColor(LIME);
    canvas.setCursor(15, 60);
    canvas.println(F("Retrying"));
    show_canvas_on_screen();
    delay(1000);
    retry();
  }
  canvas.fillScreen(BLACK);
  canvas.setTextColor(LIME);
  canvas.setCursor(3, 3);
  canvas.println(F("Connected"));
  canvas.println(F("to OBD2!"));
  show_canvas_on_screen();
  delay(1000);
  canvas.fillScreen(BLACK);
  show_canvas_on_screen();
}

//_____________________________________________________________

void brightness()
{
  bright = analogRead(pResistor);                  // readout the sensor
  lux = map(bright, MIN_LIGHT, MAX_LIGHT, 0, 255);        // map the light reading (0, 4095, 0, 255)
  if (lux < 1) {
    ledcWrite(ledChannel, 5);  // brightness value if no light is shining
  }
  else if (lux >= 1 && lux < 150) {
    ledcWrite(ledChannel, lux);  // brightness value mapping from 1 to 150
  }
  else if (lux >= 150) {
    ledcWrite(ledChannel, 255);  // brightness if #your preferred# maximum light is reached
  }
}
